/*

Copyright (c) 2011, 2012, Simon Howard

Permission to use, copy, modify, and/or distribute this software
for any purpose with or without fee is hereby granted, provided
that the above copyright notice and this permission notice appear
in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 */

#ifndef LHASA_LHA_INPUT_STREAM_H
#define LHASA_LHA_INPUT_STREAM_H

#include "lhasa.h"

/* uglify global functions */
#define lha_input_stream_read libxmp_lha_input_stream_read
#define lha_input_stream_skip libxmp_lha_input_stream_skip

/**
 * Read a block of data from the LHA stream, of the specified number
 * of bytes.
 *
 * @param stream       The input stream.
 * @param buf          Pointer to buffer in which to store read data.
 * @param buf_len      Size of buffer, in bytes.
 * @return             Non-zero if buffer was filled, or zero if an
 *                     error occurred, or end of file was reached.
 */

int lha_input_stream_read(LHAInputStream *stream, void *buf, size_t buf_len);

/**
 * Skip over the specified number of bytes.
 *
 * @param stream       The input stream.
 * @param bytes        Number of bytes to skip.
 * @return             Non-zero for success, zero for failure.
 */

int lha_input_stream_skip(LHAInputStream *stream, size_t bytes);

#endif /* #ifndef LHASA_LHA_INPUT_STREAM_H */
